#include <stdio.h>
#include <stdlib.h>
#include <mex.h>
#include <math.h>
#include <time.h>

#include "cuda.h"
#include "device_functions.h"
#include "CCMF_ML_kernels.cu"

#define YES 0
#define NO 1

#define PI 3.1415926

#define MAX(a,b) ( a > b ? a : b )
#define MIN(a,b) ( a <= b ? a : b )
#define SIGN(x) ( x >= 0.0 ? 1.0 : -1.0 )
#define ABS(x) ( (x) > 0.0 ? x : -(x) )

#ifndef HAVE_RINT 
#define rint(A) floor((A)+(((A) < 0)? -0.5 : 0.5)) 
#endif


void Check_CUDA_Error(const char *message)
{
    cudaError_t error = cudaGetLastError();
    if(error !=cudaSuccess) 
     {
       printf("ERROR: %s: %s\n", message, cudaGetErrorString(error));
     }                        
}

/**********************************************/
/************** MAIN FUNCTION *****************/
/**********************************************/

/****************************************/
extern void mexFunction(int iNbOut, mxArray *pmxOut[], int iNbIn, const mxArray *pmxIn[])
{
    
  /* iNbOut: number of outputs
     pmxOut: array of pointers to output arguments */
    
  /* iNbIn: number of inputs
     pmxIn: array of pointers to input arguments */
    
    /*    CPU Variables */
    float   *pfu, *pfCt, *pfqq, *pfcvg, *pfVecParameters, *pfpenalty;
    float   *pfbx, *pfby, *pfps, *pfpt, *pfdv, *pfgk;
    float   fError, cc, steps, beta, gamma1, gamma2;
    float   fpt, fps, *tt;
    int     *punum, iNy, iNx, iLab, iNz, iNdim, iDim[4], ix, iy, ilab, iz, id, iNI;
    int     iNbIters, szImg, SZF, index, idlab, iDev;
    float   fpts1, fpts2, fpts3, fpts4; 
    float   *fpv1, *fpv2;
    
    time_t  start_time, end_time;
    size_t fre, total;
    
    /*    GPU Variables */
    float   *pfbx_GPU, *pfby_GPU, *pfpenalty_GPU, *pfdv_GPU, *pfqq_GPU;
    float   *pfps_GPU, *pfpt_GPU, *pfgk_GPU, *pfu_GPU, *pfCt_GPU;
    float   *FPS, *FPS_GPU, FCP, FINV, FINVX, FINVY;

    /* volume ratio flow */
    float   *pfpw, *pfpw_GPU;

    /*    Output GPU information */
    cudaDeviceProp prop;
    cudaGetDeviceCount(&iDev);
    if ((unsigned int)iDev == 0)
    {
        printf("There is no CUDA device found!");
        return;
    }
    else
    {   
        printf("---------------Output GPU information------------------- \n\n");
        printf("There are %d CUDA devices in your computer. \n", iDev);
        for(int ii = 0; ii < iDev; ii ++)
        {
            cudaGetDeviceProperties(&prop, ii);
            printf("------ General Information for CUDA device %d ------ \n", ii);
            printf("Name:  %s \n", prop.name);
            printf("Multiprocessor count:  %d \n", prop.multiProcessorCount);
            printf("Total global memory: %f GB\n", float(prop.totalGlobalMem)/1024/1024/1024);
            printf("---------------------------------------------------- \n\n");
         }
    }
    
    /* Inputs */
    pfpenalty = (float*)mxGetData(pmxIn[0]);        /* penalty parameters */
    pfCt = (float*)mxGetData(pmxIn[1]);             /* bound of sink flows */
    pfVecParameters = (float*)mxGetData(pmxIn[2]);  /* Vector of parameters */

     /* 
     *pfVecParameters Setting
     * [0] : number of rows 
     * [1] : number of columns
     * [2] : number of slices
     * [3] : the maximum iteration number
     * [4] : error criterion
     * [5] : cc for the step-size of ALM
     * [6] : steps for the step-size of projected-gradient of p
     * [7] : number of labels
     * [8] : weight of the similarty terms
     * [9] : left to right lung volume ratio
     * [10] : weight of the volume ratio term
     */
    
    /* Size */
    iNy = (int) pfVecParameters[0];
    iNx = (int) pfVecParameters[1];
    iNz = (int) pfVecParameters[2];

    /* Segmentation parameters */
    iNbIters = (int) pfVecParameters[3];    /* total number of iterations */
    fError = (float) pfVecParameters[4];    /* error criterion */
    cc = (float) pfVecParameters[5];        /* cc for ALM */
    steps = (float) pfVecParameters[6];     /* steps for each iteration */
    iLab = (int) pfVecParameters[7];
    beta = (float) pfVecParameters[8];
    gamma1 = (float) pfVecParameters[9];
    gamma2 = (float) pfVecParameters[10];

    /*Some constants*/
    szImg = iNy * iNx * iNz;
    SZF = iNy*iNx;

    FCP = 1/(cc*iLab);

    printf("Initializing ................................................ \n\n");

    /* Outputs */
    /* outputs the computed u(x)  */
    iNdim = 4;
    iDim[0] = iNy;
    iDim[1] = iNx;
    iDim[2] = iNz; 
    iDim[3] = iLab;
    
    pmxOut[0] = mxCreateNumericArray(iNdim,(const int*)iDim,mxSINGLE_CLASS,mxREAL);
    pfu = (float*)mxGetData(pmxOut[0]);
   
    /* outputs the convergence rate  */
    iNdim = 2;
    iDim[0] = 1;
    iDim[1] = iNbIters;
    pmxOut[1] = mxCreateNumericArray(iNdim,(const int*)iDim,mxSINGLE_CLASS,mxREAL);
    pfcvg = (float*)mxGetData(pmxOut[1]);
    
    /* outputs the iteration number  */
    iNdim = 2;
    iDim[0] = 1;
    iDim[1] = 1;
    pmxOut[2] = mxCreateNumericArray(iNdim,(const int*)iDim,mxUINT16_CLASS,mxREAL);
    punum = (int*)mxGetData(pmxOut[2]);

    /* outputs the computation time */ 
    iNdim = 2;
    iDim[0] = 1;
    iDim[1] = 1;
    pmxOut[3] = mxCreateNumericArray(iNdim,(const int*)iDim,mxSINGLE_CLASS,mxREAL);
    tt = (float *)mxGetData(pmxOut[3]);    

    /* Memory allocation */
    /* allocate the memory for px1 */
    pfbx = (float *) calloc( (unsigned)(iNy*(iNx+1)*iLab*iNz), sizeof(float) );
    if (!pfbx)
        mexPrintf("Memory allocation failure\n");
    
    /* allocate the memory for py1 */
    pfby = (float *) calloc( (unsigned)((iNy+1)*iNx*iLab*iNz), sizeof(float) );
    if (!pfby)
        mexPrintf("Memory allocation failure\n");
    
    /* allocate the memory for ps */
    pfps = (float *) calloc( (unsigned)(iNy*iNx*iNz), sizeof(float) );
    if (!pfps)
        mexPrintf("Memory allocation failure\n");

    /* allocate the memory for pt */
    pfpt = (float *) calloc( (unsigned)(iNy*iNx*iLab*iNz), sizeof(float) );
    if (!pfpt)
        mexPrintf("Memory allocation failure\n");

    /* allocate the memory for coupled flow qq */
    pfqq = (float *) calloc( (unsigned)(iNy*iNx*iLab*(iNz)), sizeof(float) );
    if (!pfqq)
        mexPrintf("Memory allocation failure\n");
    
    /* allocate the memory for coupled flow gk */
    pfgk = (float *) calloc( (unsigned)(iNy*iNx*iLab*(iNz)), sizeof(float) );
    if (!pfgk)
        mexPrintf("Memory allocation failure\n");
    
    /* allocate the memory for divergence */
    pfdv = (float *) calloc( (unsigned)(iNy*iNx*iLab*iNz), sizeof(float) );
    if (!pfdv)
        mexPrintf("Memory allocation failure\n");
    
    /* allocate the memory for FPS */
    FPS = (float *) calloc( (unsigned)(iNy*iNx*iNz), sizeof(double) );
    if (!FPS)
        mexPrintf("Memory allocation failure\n");

    /* allocate the memory for pw */
    pfpw = (float *) calloc( (unsigned)(iLab*iNz), sizeof(float) );
    if (!pfpw)
        mexPrintf("Memory allocation failure\n");

    /* Preprocessing initial values for flows and u */
    for (iz=0; iz < iNz; iz++)
    {
        for (ix=0; ix < iNx; ix++)
        {
            /* idx = ix*iNy; */
            for (iy=0; iy < iNy; iy++)
            {
                /*------------------in the first 3D volume ------*/
                index = 0*iNx*iNy*iNz + iz*iNx*iNy + ix*iNy + iy;
                /*-----------------------------------------------*/
                fpt = pfCt[index];
                pfdv[index] = pfby[index+1] - pfby[index] 
                            + pfbx[index+iNy] - pfbx[index];
                id = 0;
                for (ilab = 1; ilab < iLab; ilab++)
                {
                    /*----------in the remaining 3D volumes ---------*/
                    idlab = ilab*iNx*iNy*iNz + iz*iNx*iNy + ix*iNy + iy;
                    /*-----------------------------------------------*/
                    pfdv[idlab] = pfby[idlab + 1] - pfby[idlab] 
                                + pfbx[idlab + iNy] - pfbx[idlab];

                    if (fpt >= pfCt[idlab])
                    {
                        fpt = pfCt[idlab];
                        id = ilab;
                    }
                }
                pfps[index] = fpt;
                pfu[index + id*iNx*iNy*iNz] = 1;

                for (ilab = 0; ilab < iLab; ilab++)
                {
                    pfpt[index + ilab*iNx*iNy*iNz] = fpt;
                }
            }
        }
    }
    
    /*    GPU Memory Allocation */
 
    cudaMemGetInfo(&fre, &total);
    mexPrintf("\nBefore allocation, total memory is: %f GB, available memory is %f GB\n", float(total)/1024/1024/1024, float(fre)/1024/1024/1024);

    cudaMalloc( (void**) &pfbx_GPU, sizeof(float)*(unsigned)((iNx+1)*iNy*iLab*iNz));
    cudaMalloc( (void**) &pfby_GPU, sizeof(float)*(unsigned)(iNx*(iNy+1)*iLab*iNz));
    cudaMalloc( (void**) &pfps_GPU, sizeof(float)*(unsigned)(iNy*iNx*iNz));
    cudaMalloc( (void**) &pfpenalty_GPU, sizeof(float)*(unsigned)(iNy*iNx*iNz));
    cudaMalloc( (void**) &pfpt_GPU, sizeof(float)*(unsigned)(iNy*iNx*iLab*iNz));
    cudaMalloc( (void**) &pfgk_GPU, sizeof(float)*(unsigned)(iNy*iNx*iLab*iNz));
    cudaMalloc( (void**) &pfdv_GPU, sizeof(float)*(unsigned)(iNy*iNx*iLab*iNz));
    cudaMalloc( (void**) &pfu_GPU, sizeof(float)*(unsigned)(iNy*iNx*iLab*iNz));
    cudaMalloc( (void**) &pfCt_GPU, sizeof(float)*(unsigned)(iNy*iNx*iLab*iNz));
    cudaMalloc( (void**) &FPS_GPU, sizeof(float)*(unsigned)(iNy*iNx*iNz));
    cudaMalloc( (void**) &pfqq_GPU, sizeof(float)*(unsigned)(iNy*iNx*iLab*(iNz)));
    cudaMalloc( (void**) &pfpw_GPU, sizeof(float)*(unsigned)(iLab*(iNz)));
   
    /*    Copy Parameters from Host to Device */

    cudaMemcpy( pfbx_GPU, pfbx, sizeof(float)*(unsigned)(iNy*(iNx+1)*iLab*iNz), cudaMemcpyHostToDevice);
    cudaMemcpy( pfby_GPU, pfby, sizeof(float)*(unsigned)((iNy+1)*iNx*iLab*iNz), cudaMemcpyHostToDevice);
    cudaMemcpy( pfps_GPU, pfps, sizeof(float)*(unsigned)(iNy*iNx*iNz), cudaMemcpyHostToDevice);
    cudaMemcpy( pfpenalty_GPU, pfpenalty, sizeof(float)*(unsigned)(iNy*iNx*iNz), cudaMemcpyHostToDevice);
    cudaMemcpy( pfpt_GPU, pfpt, sizeof(float)*(unsigned)(iNy*iNx*iLab*iNz), cudaMemcpyHostToDevice);
    cudaMemcpy( pfdv_GPU, pfdv, sizeof(float)*(unsigned)(iNy*iNx*iLab*iNz), cudaMemcpyHostToDevice);
    cudaMemcpy( pfu_GPU, pfu, sizeof(float)*(unsigned)(iNy*iNx*iLab*iNz), cudaMemcpyHostToDevice);
    cudaMemcpy( pfCt_GPU, pfCt, sizeof(float)*(unsigned)(iNy*iNx*iLab*iNz), cudaMemcpyHostToDevice);
    cudaMemcpy( FPS_GPU, FPS, sizeof(float)*(unsigned)(iNy*iNx*iNz), cudaMemcpyHostToDevice);
    cudaMemcpy( pfqq_GPU, pfqq, sizeof(float)*(unsigned)(iNy*iNx*iLab*(iNz)), cudaMemcpyHostToDevice);
    cudaMemcpy( pfpw_GPU, pfpw, sizeof(float)*(unsigned)(iLab*(iNz)), cudaMemcpyHostToDevice);

    /* Compute the execution configuration */
    cudaMemGetInfo(&fre, &total);
    mexPrintf("\nAfter allocation, total memory is: %f GB, available memory is %f GB\n", float(total)/1024/1024/1024, float(fre)/1024/1024/1024);
    
   /*   GPU blocks and grids configuration*/
   dim3 dimBlock(BLOCK_SIZE,BLOCK_SIZE,BLOCK_SIZE);
  
   /**************************** dimGrid*******************************/
   int blocksInX = (iNy/dimBlock.x) + (!(iNy%dimBlock.x)?0:1);
   int blocksInY = (iNx/dimBlock.y) + (!(iNx%dimBlock.y)?0:1);
   int blocksInZ = (iNz/dimBlock.z) + (!(iNz%dimBlock.z)?0:1);
   //mexPrintf("blocksInX is: %d, blocksInY is: %d, blocksInZ is: %d \n", blocksInX, blocksInY, blocksInZ);
   dim3 dimGrid ( blocksInX, blocksInY*blocksInZ);

   /**************************** dimGrid_x*******************************/
   blocksInX = ((iNy-1)/dimBlock.x) + (!((iNy-1)%dimBlock.x)?0:1);
   int blocksInY_x = (iNx/dimBlock.y) + (!(iNx%dimBlock.y)?0:1);
   blocksInZ = (iNz/dimBlock.z) + (!(iNz%dimBlock.z)?0:1);
   //mexPrintf("blocksInX is: %d, blocksInY_x is: %d, blocksInZ is: %d \n", blocksInX, blocksInY_x, blocksInZ);
   dim3 dimGrid_x (blocksInX, blocksInY_x*blocksInZ);

   /**************************** dimGrid_y*******************************/
   blocksInX = (iNy/dimBlock.x) + (!(iNy%dimBlock.x)?0:1);
   int blocksInY_y = ((iNx-1)/dimBlock.y) + (!((iNx-1)%dimBlock.y)?0:1);
   blocksInZ = (iNz/dimBlock.z) + (!(iNz%dimBlock.z)?0:1);
   //mexPrintf("blocksInX is: %d, blocksInY_y is: %d, blocksInZ is: %d \n", blocksInX, blocksInY_y, blocksInZ);
   dim3 dimGrid_y (blocksInX, blocksInY_y*blocksInZ);

   /**************************** dimGrid_z*******************************/
   blocksInX = (iNy/dimBlock.x) + (!(iNy%dimBlock.x)?0:1);
   int blocksInY_z = (iNx/dimBlock.y) + (!(iNx%dimBlock.y)?0:1);
   blocksInZ = ((iNz-1)/dimBlock.z) + (!((iNz-1)%dimBlock.z)?0:1);
   //mexPrintf("blocksInX is: %d, blocksInY_z is: %d, blocksInZ is: %d \n", blocksInX, blocksInY_z, blocksInZ);
   dim3 dimGrid_z (blocksInX, blocksInY_z*blocksInZ);

   /**************************** dimGrid*******************************
   dim3 dimBlock2d(BLOCK2D_SIZE,BLOCK2D_SIZE);
   dim3 dimGrid2d ( (iNy/dimBlock2d.x) + (!(iNy%dimBlock2d.x)?0:1) ,
                  (iNx/dimBlock2d.y) + (!(iNx%dimBlock2d.y)?0:1) );
   */

   /**************************** dimGrid3d*****************************
   dim3 dimBlock3d(BLOCK2D_SIZE, BLOCK2D_SIZE, BLOCK2D_SIZE);
   dim3 dimGrid3d ( (iNy/dimBlock3d.x) + (!(iNy%dimBlock3d.x)?0:1) ,
                    (iNx/dimBlock3d.y) + (!(iNx%dimBlock3d.y)?0:1) ,
                    (iNz/dimBlock3d.z) + (!(iNz%dimBlock3d.z)?0:1));
   */

    /* define the constants (these were problematic and were not used Fumin's implementation) */
    FINV = 1.0/(double)blocksInY;
    FINVY = 1.0/(double)blocksInY_y;
    FINVX = 1.0/(double)blocksInY_x;

    /*  Main iterations */
  
    iNI = 0;
    start_time = clock();
    
    printf("Start computing ......................................... \n\n");

    while( iNI<iNbIters ) 
    { 
        
        /* update p */
        krnl_1<<<dimGrid, dimBlock>>>(pfdv_GPU, pfpt_GPU, pfps_GPU, pfu_GPU, pfgk_GPU, cc, 
                                      iNx, iNy, iLab, iNz, SZF, szImg, blocksInY, FINV, pfqq_GPU,pfpw_GPU);

        krnl_2<<<dimGrid_y, dimBlock>>>(pfbx_GPU, pfgk_GPU, steps, 
                                      iNx, iNy, iLab, iNz, SZF, szImg, blocksInY_y, FINVY);

        krnl_3<<<dimGrid_x, dimBlock>>>(pfby_GPU, pfgk_GPU, steps, 
                                      iNx, iNy, iLab, iNz, SZF, szImg, blocksInY_x, FINVX);

        krnl_4<<<dimGrid, dimBlock>>>(pfbx_GPU, pfby_GPU, pfgk_GPU, pfpenalty_GPU, 
                                      iNx, iNy, iLab, iNz, SZF, szImg, blocksInY, FINV);

        krnl_5<<<dimGrid_y, dimBlock>>>(pfbx_GPU, pfgk_GPU, 
                                      iNx, iNy, iLab, iNz, SZF, szImg, blocksInY_y, FINVY);

        krnl_6<<< dimGrid_x, dimBlock>>>(pfby_GPU, pfgk_GPU, 
                                      iNx, iNy, iLab, iNz, SZF, szImg, blocksInY_x, FINVX);

        krnl_7<<<dimGrid, dimBlock>>>(pfbx_GPU, pfby_GPU, pfdv_GPU, pfps_GPU, pfpt_GPU, pfu_GPU, 
                                      pfCt_GPU, pfgk_GPU, cc, 
                                      iNx, iNy, iLab, iNz, SZF, szImg, blocksInY, FINV, beta, pfqq_GPU, pfpw_GPU);
        
        krnl_qq<<<dimGrid, dimBlock>>>(pfbx_GPU, pfby_GPU, pfdv_GPU, pfps_GPU, pfpt_GPU,
                                       pfu_GPU,  pfCt_GPU, pfgk_GPU, cc,
                                      iNx, iNy, iLab, iNz, SZF, szImg, blocksInY, FINV, pfqq_GPU, beta, pfpw_GPU);

        krnl_8<<< dimGrid, dimBlock>>>(pfdv_GPU, pfps_GPU, pfpt_GPU, pfu_GPU, pfgk_GPU, FPS_GPU, cc,
                                      iNx, iNy, iLab, iNz, SZF, szImg, blocksInY, FINV, FCP, pfqq_GPU, pfpw_GPU);   

        cudaMemcpy(pfgk, pfgk_GPU, sizeof(float)*(unsigned)(iNy*iNx*iLab*iNz), cudaMemcpyDeviceToHost);


        /******************************/

        for (int idz=0; idz< iNz; idz++)
        {
            fpts1 = 0;
            fpts2 = 0;
            fpts3 = 0;
            fpts4 = 0;
            
    
            for (int idx=0; idx < iNy; idx++)
            {
                for (int idy=0; idy < iNx; idy++)
                {
                    int index = idx +  idy*iNy + idz*SZF;
                    
                    fpts1 += pfgk[index] - gamma1*fpv2[idz];
                    fpts2 += pfgk[index+szImg] + fpv2[idz];
                    
                    fpts3 += pfgk[index] + gamma2*fpv1[idz];
                    fpts4 += pfgk[index+szImg] - fpv1[idz];
                }
            }          

            fpv1[idz] = (fpts2/SZF - gamma2*fpts1/SZF) / (1 + gamma2*gamma2);
            fpv1[idz] = MAX(fpv1[idz], 0);

            
            fpv2[idz] = (gamma1*fpts3/SZF - fpts4/SZF) / (1 + gamma1*gamma1);
            fpv2[idz] = MAX(fpv2[idz], 0);

            pfpw[idz*iLab + 0] = gamma2*fpv1[idz] - gamma1*fpv2[idz]; 
            pfpw[idz*iLab + 1] = fpv2[idz] - fpv1[idz]; 
            pfpw[idz*iLab + 2] = 0;
        }
        /******************************/
        cudaMemcpy(pfpw_GPU, pfpw, sizeof(float)*(unsigned)(iLab*(iNz)), cudaMemcpyHostToDevice);

        krnl_9<<< dimGrid, dimBlock>>>(pfdv_GPU, pfps_GPU, pfpt_GPU, pfu_GPU, pfgk_GPU, FPS_GPU, cc,
                                      iNx, iNy, iLab, iNz, SZF, szImg, blocksInY, FINV, FCP, pfqq_GPU, pfpw_GPU);   
    
        cudaMemcpy(FPS, FPS_GPU, sizeof(float)*(unsigned)(iNx*iNy*iNz), cudaMemcpyDeviceToHost);
        
        
        fps = 0;
        for (int ii=0; ii< iNx*iNy*iNz; ii++)
        {
                fps += FPS[ii];
        }

        pfcvg[iNI] = fps / (iNx*iNy*iLab*iNz);
        
        if (pfcvg[iNI] <= fError)
            break;

        iNI ++;
        
        /*
        float *pError;
        pError = (float *) calloc( (unsigned)(iNz), sizeof(float) );
        if (!pError)
        mexPrintf("Memory for error allocation failure\n"); 
        bool exit = 1;

        for (int jj = 0; jj < iNz; jj++)
        {
            fps = 0;
            for (int ii=iNx*iNy*jj; ii< iNx*iNy*(jj+1); ii++)
                {
                    fps += FPS[ii];
                }
            pError[jj] = fps / (iNx*iNy*iLab);
            if(pError[jj] >= fError)
                {
                    exit = 0;
                }
            pfcvg[iNI] += pError[jj];
        }
        
        pfcvg[iNI] = pfcvg[iNI]/iNz;
        
        if (pfcvg[iNI] <= fError && exit ==1)
            break;

        iNI ++;
        */
     } 
    
    /*  Copy GPU segmentation results back to CPU*/
    cudaMemcpy(pfu, pfu_GPU, sizeof(float)*(unsigned)(iNy*iNx*iNz*iLab), cudaMemcpyDeviceToHost);
    end_time = clock();

    /* Outputs (see above) */
    punum[0] = ++iNI;
    mexPrintf("number of iterations = %i\n",iNI);
    
    /* Free memory */
    free( (float *) pfbx );
    free( (float *) pfby );
    free( (float *) pfps );
    free( (float *) pfpt );
    free( (float *) pfdv );
    free( (float *) FPS );
    free( (float *) pfqq );
    free( (float *) pfpw );

    /* Free GPU Memory */
    cudaFree(pfbx_GPU);
    cudaFree(pfby_GPU);
    cudaFree(pfps_GPU);
    cudaFree(pfpenalty_GPU);
    cudaFree(pfpt_GPU);
    cudaFree(pfgk_GPU);
    cudaFree(pfdv_GPU);
    cudaFree(pfu_GPU);
    cudaFree(pfCt_GPU);
    cudaFree(FPS_GPU);
    cudaFree(pfqq_GPU);
    cudaFree(pfpw_GPU);
 
    tt[0] = difftime(end_time, start_time)/1000000;
    mexPrintf("\nComputing Time for max-flow = %.4f sec\n", tt[0]);  
    
    cudaMemGetInfo(&fre,&total);
    mexPrintf("\nAfter free these variables, total memory is: %f GB, free memory is %f GB\n", float(total)/1024/1024/1024, float(fre)/1024/1024/1024);
}
/****************************************/






/**********************************************/
/************** END MAIN FUNCTION *************/
/**********************************************/
